'''
  $Id$

  clazzes.org PBA package downloader

  (c)2018 clazzes.org project
  
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at
 
      http://www.apache.org/licenses/LICENSE-2.0
 
  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.

'''

import logging
import re
import io
import os.path
import glob
import bz2
import gzip
import tempfile
import shutil
import subprocess

from collections import defaultdict

from urllib.request import urlopen
from urllib.error import HTTPError

from clazzes.pbadownload import util

log = logging.getLogger(__name__)

KNOWN_HASHES = ('MD5Sum','MD5sum','SHA1','SHA256','SHA512')

class Release(util.RepoDataSupport):
    '''
      The representation of a debian Release file.
    '''
    
    __slots__ = ("codename","suite","origin","label","version","architectures","components","description","date","checksums")
    
    
    def __init__(self,fh,source):
        '''
           Parse a releases file from a StringIO file-like object.
           :param fh: a file-like handlereturn strings.
           :param source: A description of the file handle like an URL or a file name.
        '''
        
        cur_checksum = None
        cur_checksum_algo = None
        checksums = None
        
        line = fh.readline()
        
        pgp_file = line == '-----BEGIN PGP SIGNED MESSAGE-----\n'
        
        # skip PGP header
        if pgp_file:
            
            while line != '\n' and line != '':
                line = fh.readline()
        
            if line != '':
                line = fh.readline()
        
        while True:
            
            # EOF
            if len(line) == 0:
                break
            
            if pgp_file and line == '-----BEGIN PGP SIGNATURE-----\n':
                break
            
            if line[-1:] == '\n':
                line = line[:-1]

            if len(line) == 0:
                continue
            
            if cur_checksum is not None:
                
                if line[0] == ' ':
                
                    # hash value, file length, filename
                    hv,fl,fn = re.split(' +',line.strip(),2)
                    cur_checksum[fn] = util.Checksum(value=hv,algorithm=cur_checksum_algo,size=int(fl))
                else:
                    cur_checksum = None
                    cur_checksum_algo = None
        
            if cur_checksum is None:
            
                kv = line.split(':',1)
            
                if len(kv) < 2:
                    log.warn("Ignoring invalid Release line [%s] of resource [%s]."%(line,source))
                    continue
                
                k = kv[0].strip()
                v = kv[1].strip()
            
                if k == 'Origin':
                    self.origin = v
                elif k == 'Label':
                    self.label = v
                elif k == 'Codename':
                    self.codename = v
                elif k == 'Suite':
                    self.suite = v
                elif k == 'Version':
                    self.version = v
                elif k == 'Description':
                    self.description = v
                elif k == 'Architectures':
                    self.architectures = v.split(' ')
                elif k == 'Components':
                    self.components = v.split(' ')
                elif k == 'Date':
                    self.date = util.parseRFC822Date(v)
                    
                elif k in KNOWN_HASHES:
                    cur_checksum = {}
                    if checksums is None:
                        checksums = {}
                    
                    checksums[k] = cur_checksum
                    
                    if k[0:3] == 'MD5':
                        cur_checksum_algo = 'MD5'
                    else:
                        cur_checksum_algo = k
                else:
                    log.warn("Ignoring unknown Release tag [%s] of resource [%s]."%(line,source))
            
            line = fh.readline()
            
        self.checksums = checksums
        self.fillMissingSlots()
        
class Package(util.RepoDataSupport):
    
    __slots__ = ("package",'version','architecture','maintainer','filename','description','section','priority','installed_size','checksums','extra')
    
    '''
      The representation of a single debian Package file stanza.
    '''
    def __init__(self,fh,source):
        '''
           Parse a Packages file stanza from a StringIO file-like object.
           :param fh: a file-like handle, which returns strings.
           :param source: A description of the file handle like an URL or a file name.
        '''
        
        self.checksums = {}
        self.extra = {}
        size = None
        
        line = fh.readline()
        
        while True:
            
            if line[-1:] == '\n':
                line = line[:-1]
            
            if len(line) == 0:
                break
            
            kv = line.split(':',1)
        
            if len(kv) < 2:
                log.warn("Ignoring invalid Packages line [%s] of resource [%s]."%(line,source))
                continue
            
            k = kv[0].strip()
            v = kv[1].strip()
       
            nextline = fh.readline()
            
            # continuation lines.
            while nextline[0:1] == ' ':
                
                if nextline[-1:] == '\n':
                    nextline = nextline[:-1]
                
                v += '\n' + nextline[1:]
                
                nextline = fh.readline()
            
            if k == 'Package':
                self.package = v
            elif k == 'Version':
                self.version  = v
            elif k == 'Architecture':
                self.architecture  = v
            elif k == 'Maintainer':
                self.maintainer = v
            elif k == 'Filename':
                self.filename  = v
            elif k == 'Size':
                size = int(v)
            elif k == 'Description':
                self.description = v
            elif k == 'Section':
                self.section = v
            elif k == 'Priority':
                self.priority = v
                
            elif k in KNOWN_HASHES:
                
                if k[0:3] == 'MD5':
                    algo = 'MD5'
                else:
                    algo = k
                
                self.checksums[k] = util.Checksum(value=v,algorithm=algo,size=size)
                
            else:
                self.extra[k] = v
            
            line = nextline
            
        self.fillMissingSlots()


def readPackages(fh,source):
    '''
       Parse a Packages file from a StringIO file-like object.
       :param fh: a file-like handle, which returns strings.
       :param source: A description of the file handle like an URL or a file name.
       :return: a dictionary mapping package names to lists of Package objects.
    '''
    packages = defaultdict(list)
    
    pkg = Package(fh,source)
    
    while pkg.package is not None:
        packages[pkg.package].append(pkg)
        pkg = Package(fh,source)
    
    log.info("Successfully loaded [%d] package headers from [%s]."%(len(packages),source))
    
    return packages


class DebRepo:
    
    __slots__ = ( "baseurl","suite","components","release","preferred_hash")
    
    def openUrl(self,relurl):
        '''
           Open an URL relative to the base URL of the repository.
           
           :param relurl: A relive URL not starting with a slash ant not containing '://' 
           :return: A file-like object as returned by urllib.request.urlopen().
        '''
        
        if relurl[0] == '/' or '://' in relurl:
            raise ValueError("Relative URL [%s] is absolute or contains a scheme.")
        
        url = self.baseurl + relurl
        
        log.info("Loading URL [%s]..."%url)
        
        return urlopen(url)
    

    def __init__(self,debline,preferred_hash="SHA256",check_signature=True,distro=None):
        '''
           Initialize a debian repository from debian repository line like
           'deb http://deb.clazzes.org/any any-dojo-1 main'.
           All parts starting from the fourth are components to load, if no component
           is given, the components are taken from the Release file inside the 
           repositories main directory.
           :param distro: A replacement for the $distro placeholder in the third componentn of
                          debline (ake the "suite").
           :param check_signature: Whether to check the GPG signature on the InRelease file.
           :param preferred_hash: The preferred hash algorithm.
        '''
        
        self.preferred_hash = preferred_hash
        
        parts = re.split(' +',debline.strip())
        if parts[0] != 'deb':
            raise ValueError("Debian repo line [%s] does not start with 'deb'."%debline)
        
        if len(parts) < 3:
            raise ValueError("Debian repo line [%s] has less than 3 parts."%debline)
        
        baseurl = parts[1]
        
        if baseurl[-1] == '/':
            self.baseurl = baseurl
        else:
            self.baseurl = baseurl + '/'

        if distro is None:
            self.suite = parts[2]
        else:
            self.suite = parts[2].replace("$distro",distro)
        
        if len(parts) > 3:
            self.components = parts[3:]
        
        url = "dists/"+self.suite+"/Release"
        
        if check_signature:
        
            with tempfile.TemporaryDirectory("pbadownload") as tmp:
                
                inrelfn = os.path.join(tmp,"InRelease")
                inurl = "dists/"+self.suite+"/InRelease"
                
                sigfn = None
                relfn = None
                
                try:
                    with self.openUrl(inurl) as fh, io.open(inrelfn,"wb") as oh:
                        shutil.copyfileobj(fh,oh)
                
                    relfn = inrelfn
                    url = inurl
                
                except HTTPError as e:
                    if e.code == 404:
                        log.warn("File [%s] not found, using Release and Release.gpg"%inurl)
                    else:
                        raise
                
                if relfn is None:
                
                    relfn = os.path.join(tmp,"Release")
                    sigfn = os.path.join(tmp,"Release.gpg")
                                 
                    with self.openUrl(url) as fh, io.open(relfn,"wb") as oh:
                        shutil.copyfileobj(fh,oh)
                        
                    with self.openUrl(url+".gpg") as fh, io.open(sigfn,"wb") as oh:
                        shutil.copyfileobj(fh,oh)
                
                # gpgv --keyring /etc/apt/trusted.gpg Release.gpg Release
                log.info("Checking GPG signature on [%s]"%url)
                
                gpgv = ["/usr/bin/gpgv"]

                file = "/etc/apt/trusted.gpg"
                
                if os.path.exists(file):
                    gpgv.append('--keyring')
                    gpgv.append(file)

                for file in glob.glob('/etc/apt/trusted.gpg.d/*.gpg'):
                    gpgv.append('--keyring')
                    gpgv.append(file)

                if sigfn is not None:
                    gpgv.append(sigfn)
                    
                gpgv.append(relfn)
                
                if log.getEffectiveLevel() <= logging.DEBUG:
                    log.debug("Running %s"%gpgv)
                                
                res = subprocess.run(gpgv,
                                stdout=subprocess.PIPE,stderr=subprocess.STDOUT,
                                universal_newlines=True)
                
                for line in res.stdout.split('\n'):
                    if len(line) > 0:
                        log.info(line)
                
                if res.returncode != 0:
                    raise util.WrongSignature("Signature check on [%s] failed with error [%d]"%(url,res.returncode))
                    
                log.info("GPG signature on [%s] is OK."%url)
        
                with io.open(relfn,"r",encoding='utf-8') as fh:
                    self.release = Release(fh,self.baseurl+url)
                
        else:
            with self.openUrl(url) as fh:
                self.release = Release(io.TextIOWrapper(fh,encoding='utf-8'),self.baseurl+url)
        
        if len(parts) == 3:
            self.components = self.release.components
            log.info("Taking components %s from [%s]"%(self.components,self.baseurl+url))
    
    def resolveArchitecture(self,arch):
        '''
           Resolve the files of the given architecture.
           return a list of pairs of a component name and a Packages object.
        '''
        filelist = self.release.checksums[self.preferred_hash]
        
        ret = []
        
        for component in self.components:
            
            basefn = component+'/binary-'+arch+'/Packages'
            found = False
            
            for compression in ('.bz2','.gz',''):
                fn = basefn + compression
            
                checksum = filelist.get(fn)
                
                if checksum is not None:
            
                    url = 'dists/'+self.suite+'/'+fn
            
                    with self.openUrl(url) as fh:
                        data = fh.read()
                    
                    checksum.checkData(data,self.baseurl+url)
                    
                    if compression == '.bz2':
                        
                        with io.TextIOWrapper(bz2.BZ2File(io.BytesIO(data)),encoding='utf-8') as ph:
                            packages = readPackages(ph,self.baseurl+url)
                            
                    elif compression == '.gz':
                        
                        with io.TextIOWrapper(gzip.GzipFile(filename=fn,fileobj=io.BytesIO(data)),encoding='utf-8') as ph:
                            packages = readPackages(ph,self.baseurl+url)
                    else:
                        
                        with io.TextIOWrapper(io.BytesIO(data),encoding='utf-8') as ph:
                            packages = readPackages(ph,self.baseurl+url)
        
                    ret.append((component,packages))
                    found = True
                    break
            
            if not found:
                raise FileNotFoundError("Cannot find [%s] or bz2 of gz variant."%basefn)
            
        return ret
