'''
  $Id$

  clazzes.org PBA package downloader

  (c)2018 clazzes.org project
  
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at
 
      http://www.apache.org/licenses/LICENSE-2.0
 
  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.

'''

import os
import sys
import contextlib
import getopt
import logging
import configparser
import tempfile
import zipfile
import apt_pkg
import rpm

apt_pkg.init_system()

from datetime import datetime
from functools import cmp_to_key

from clazzes.pbadownload import util
from clazzes.pbadownload import yum
from clazzes.pbadownload import deb

log = logging.getLogger(__name__)

def usage(argv0):
    print("Usage: %s [options...] [<pkg1> [<pkg2> ...]]"%argv0,file=sys.stderr)
    print("  Supported options are:",file=sys.stderr)
    print("    --debug   Activate debug logging.",file=sys.stderr)
    print("    --help|-h Print this help test.",file=sys.stderr)
    print("    -a <arch> Specify the architecture of the packages to download.",file=sys.stderr)
    print("    -d <distro> Specify the distro of the packages to download for yum or deb repos.",file=sys.stderr)
    print("                The distro is substituted for $relversion in yum repo URLs and.",file=sys.stderr)
    print("                for $distro in the suite part of deb lines.",file=sys.stderr)
    print("    -y <repourl> Specify a yum repo url to use, multiple repos may be specified",file=sys.stderr)
    print("                 by adding multiple -y options.",file=sys.stderr)
    print("                 YUM repo URLs may bes sspecifed as an URL of a repo file like",file=sys.stderr)
    print("                   http://rpm.clazzes.org/repos/fancylibs-3.7/fancylibs-3.7.repo",file=sys.stderr)
    print("                 or an URL with a trailing slash poitning to a parametrized directory like.",file=sys.stderr)
    print("                   https://rpm.clazzes.org/repo/fancylibs-3.7/centos$releasever/$basearch/",file=sys.stderr)
    print("    -r <debline> Specify a deb repo line to use, multiple repos may be specified",file=sys.stderr)
    print("                 by adding multiple -d options.",file=sys.stderr)
    print("    -c <config> Specify a config file to read instad of cmdline params.",file=sys.stderr)
    print("    -z <zipfile> Download to a temporary folder and ZIP the contents to <zipfile>.",file=sys.stderr)
    print("    -D <dir> Download to the directory <dir> instead to the current directory.",file=sys.stderr)
    print("             Only used, is -z is not given.",file=sys.stderr)
    print("    -p <policy> specifies policy for resolving different version of a package,")
    print("                can have the value \"first or \"newest\" defaults to \"first\".")
    print("    --nocheck|-k Don't check debian archiver signatures.",file=sys.stderr)


def run_downloader(argv0,args):
    '''
      The main program for downloading files
      :param argv0: The name under which the program has been invoked.
      :param args: Command line arguments 
    '''
    
    arch = None
    distro = None
    zipfn = None
    targetdir = None
    check_signature = True
    yum_repo_urls = []
    deb_repo_lines = []
    policy = "first"
    
    try:
        opts, rawpackages = getopt.getopt(args, "a:d:y:r:z:D:p:c:kh", ["arch=","distro=","yum=","repo=","zip=","dir=","policy=","config=","nocheck","help"])
    except getopt.GetoptError as err:
        # print help information and exit:
        print(err,file=sys.stderr)  # will print something like "option -a not recognized"
        usage(sys.argv[0])
        sys.exit(64)
    
    # pairs of package name / target directory.
    packages = []
    
    for rawpkg in rawpackages:
        i = rawpkg.rfind('/')
        if i < 0:
            packages.append((rawpkg,None))
        else:
            packages.append((rawpkg[i+1:],rawpkg[0:i]))
    
    for o, a in opts:
        if o in ("-h,--help"):
            usage(sys.argv[0])
            sys.exit(0)
        elif o in ("-a", "--arch"):
            arch = a
        elif o in ("-d", "--distro"):
            distro = a
        elif o in ("-y", "--yum"):
            yum_repo_urls.append(("cmdline_%d"%len(yum_repo_urls),a))
        elif o in ("-r", "--repo"):
            deb_repo_lines.append(a)
        elif o in ("-z", "--zip"):
            zipfn = a
        elif o in ("-D", "--dir"):
            targetdir = a
        elif o in ("-p", "--policy"):
            policy = a
        elif o in ("-k", "--nocheck"):
            check_signature = False
        elif o in ("-c", "--config"):
            
            try :
                conf = configparser.ConfigParser(allow_no_value=True)
                conf.optionxform = lambda option: option
                
                log.info("Loading config file [%s]."%a)
                conf.read(a,'utf-8')
            
                if conf.has_section('options'):
                    arch = conf['options'].get('arch',arch)
                    distro = conf['options'].get('distro',distro)
                    check_signature = conf['options'].getboolean('check_signature',check_signature)
                    zipfn = conf['options'].get('zip',zipfn)
                    targetdir = conf['options'].get('dir',targetdir)
                    policy = conf["options"].get("policy", policy)
 
                if conf.has_section('packages'):
                    for k,v in conf.items('packages'):
                        packages.append((k,v))
            
                if conf.has_section('yum'):
                    for k,v in conf.items('yum'):
                        log.info("Adding YUM repo [%s] with URL [%s]."%(k,v))
                        yum_repo_urls.append((k,v))

                if conf.has_section('deb'):
                    for k,v in conf.items('deb'):
                        log.info("Adding DEB repo [%s] with source [%s]."%(k,v))
                        deb_repo_lines.append(v)
                        
                for sec in conf.sections():
                    if sec[-9:] == '-packages':
                        for k,v in conf.items(sec):
                            if v is None:
                                packages.append((k,sec[0:-9]))
                            else:
                                packages.append((k,os.path.join(sec[0:-9],v)))
                        
            except:
                log.exception("%s: Error loading config file [%s]"%(argv0,a))
                sys.exit(64)
        else:
            assert False, "unhandled option"

    if len(packages) == 0:
        print("Nothing to download, specify at least one package.",file=sys.stderr)
        usage(sys.argv[0])
        sys.exit(2)
        
    subdirs = set()
    
    for pkg,subdir in packages:  # @UnusedVariable
        if subdir is not None:
            subdirs.add(subdir)
        
    if targetdir is not None and zipfn is not None:
        log.warn("Ignoring target directory [%s] in favor of zip output file [%s]"%(targetdir,zipfn))

    if len(yum_repo_urls) > 0:
        
        if len(deb_repo_lines) > 0:
            print("You cannot mix debian and yum repositories.",file=sys.stderr)
            usage(sys.argv[0])
            sys.exit(64)
        
        if arch is None:
            print("No arch specified for yum repositories, use option -a.",file=sys.stderr)
            usage(sys.argv[0])
            sys.exit(64)
    
        if distro is None:
            print("No distro specified for yum repositories, use option -d.",file=sys.stderr)
            usage(sys.argv[0])
            sys.exit(64)
    
        try:
            
            if zipfn is None:
            
                mkdirs(targetdir,subdirs)
                files = download_yum(yum_repo_urls,arch,distro,packages,targetdir,policy)
                log.info("Successfully downloaded [%d] RPM files."%len(files))
                
            else:
                with tempfile.TemporaryDirectory("pbadownload") as tmp:
                    
                    mkdirs(tmp,subdirs)
                    files = download_yum(yum_repo_urls,arch,distro,packages,tmp,policy)
                    writeZipFile(zipfn,tmp,files)
                
        except:
            log.exception("%s: Caught exception downloading yum repositories"%argv0)
            sys.exit(1)
    
    elif len(deb_repo_lines) > 0:
        
        if arch is None:
            print("No arch specified for deb repositories, use option -a.",file=sys.stderr)
            usage(sys.argv[0])
            sys.exit(64)

        try:
            if zipfn is None:
            
                mkdirs(targetdir,subdirs)
                files = download_deb(deb_repo_lines,arch,distro,check_signature,packages,targetdir,policy)
                log.info("Successfully downloaded [%d] DEB files."%len(files))
                              
            else:
                with tempfile.TemporaryDirectory("pbadownload") as tmp:
                    
                    mkdirs(tmp,subdirs)
                    files = download_deb(deb_repo_lines,arch,distro,check_signature,packages,tmp,policy)
                    writeZipFile(zipfn,tmp,files)
        except:
            log.exception("%s: Caught exception downloading deb repositories"%argv0)
            sys.exit(1)

    else:
        if arch is None:
            print("No repository specified, use option -y or -r.",file=sys.stderr)
            usage(sys.argv[0])
            sys.exit(64)

def mkdirs(tmp,subdirs):
    for subdir in subdirs:
        if tmp is None:
            os.makedirs(subdir,exist_ok=True)
        else:
            os.makedirs(os.path.join(tmp,subdir),exist_ok=True)

def writeZipFile(zipfn,tmp,files):
    
    log.info("Writing [%d] downloaded files to ZIP file [%s]..."%(len(files),zipfn))
    
    with zipfile.ZipFile(zipfn,mode="w") as zipfh:
        
        for rfn in files:
            
            fn = os.path.join(tmp,rfn)
            
            log.info("Compressing file [%s] to ZIP file."%rfn)
            zipfh.write(fn,rfn)
            
    log.info("Successfully wrote [%d] downloaded files to ZIP file [%s]."%(len(files),zipfn))

class RPMPolicy:
    def __new__(cls, name):
        return {"first": cls.first, "newest": cls.newest}[name]
    @staticmethod
    def first(l):
        return l[0]
    @staticmethod
    def newest(l):

        k = cmp_to_key(rpm.labelCompare)
        return max(l, key=lambda p: k((p[1].epoch, p[1].version, p[1].release)))

def download_yum(yum_repo_urls,arch,distro,packages,targetdir,policy):

    policy = RPMPolicy(policy)
    
    # open yum databases
    yum_repos = []
    
    files = []

    # load repo files.
    for name,url in yum_repo_urls:
        yum_repos.append(yum.YumRepo(name,url))

    with contextlib.ExitStack() as stack:
    
        yum_dbs = {}
    
        for yum_repo in yum_repos:
            
            baseurl = yum_repo.resolveBaseUrl(basearch=arch,releasever=distro)
            
            repomd = yum.RepoMd(baseurl)
            
            log.info("Open DB for yum repo [%s]"%yum_repo.name)
            
            db = stack.enter_context(repomd.openDb())

            yum_dbs[yum_repo.name] = (repomd,db)

        for pkg,subdir in packages:

            candidates = []

            for yum_repo in yum_repos:
            
                name = yum_repo.name    
                repomd,db = yum_dbs[name]
                
                packages = db.findPackage(pkg)

                for package in packages:
                    candidates.append((repomd, package))
                    log.info("Found package [%s] in repository [%s]"%(pkg,name))
            
            if len(candidates) == 0:
                raise FileNotFoundError("Package [%s] not found in any repository."%pkg)

            repomd, info = policy(candidates)
            
            rfn = info.location[info.location.rfind('/')+1:]
            
            if subdir is not None:
                rfn = os.path.join(subdir,rfn)
            
            if targetdir is None:
                fn = rfn
            else:
                fn = os.path.join(targetdir,rfn)
            
            with repomd.openUrl(info.location) as fh:
                
                lm = fh.info().get("Last-Modified")
                
                hd = info.checksum.saveAndCheckFile(fh,fn)
                
            log.info("Saved file [%s] with checksum [%s]"%(rfn,hd))

            if lm is not None:
                
                dt = util.parseRFC822Date(lm)
                log.info("Setting modification time of file [%s] to [%s]"%(rfn,dt))
                os.utime(fn,(datetime.now().timestamp(),dt.timestamp()))

            files.append(rfn)
            
    return files

class DebPolicy(object):
    def __new__(cls, name):
        return {"first": cls.first, "newest": cls.newest}[name]
    @staticmethod
    def first(l):
        return l[0]
    @staticmethod
    def newest(l):
        k = cmp_to_key(apt_pkg.version_compare)
        return max(l, key=lambda p: k(p[1].version))

def download_deb(deb_repo_lines,arch,distro,check_signature,packages,targetdir,policy):

    policy = DebPolicy(policy)
    
    # openload deb repos
    deb_repos = []
    
    files = []

    # load repo files.
    for debline in deb_repo_lines:
        
        repo = deb.DebRepo(debline,distro=distro,check_signature=check_signature)
        components = repo.resolveArchitecture(arch)
        
        deb_repos.append((repo,components))
        
    for pkg,subdir in packages:

        candidates = []

        for deb_repo in deb_repos:
            
            repo,components = deb_repo
            
            name = repo.suite    
            
            for cp in components:
                
                component,packages = cp

                pkglist = packages.get(pkg)

                if pkglist:
                    for pkgobj in pkglist:
                        candidates.append((repo,pkgobj))
                    
                        log.info("Found package [%s] in repository [%s/%s] with version [%s]"%(pkg,name,component,pkgobj.version))
            
        if len(candidates) == 0:
            raise FileNotFoundError("Package [%s] not found in any repository."%pkg)

        repo, info = policy(candidates)
            
        rfn = info.filename[info.filename.rfind('/')+1:]
            
        '''
          remove "illegal" jessie_, stretch before arch (amd64, i386, armhf, all),
          to satisfy standards and ansible
        '''
        if log.getEffectiveLevel() <= logging.DEBUG:
            log.debug("Original fn: [%s]" % rfn)
            
        for arch in ["amd64", "i386", "armhf", "all"]:
            for distro in ["wheezy", "jessie", "stretch", "buster", "any"]:
                
                repl = "_" + distro + "_" + arch + ".deb"
                
                if rfn.endswith(repl):
                    fixed_fn = rfn[:-len(repl)] + "_" + arch + ".deb"
                    rfn = fixed_fn
        
        if log.getEffectiveLevel() <= logging.DEBUG:
            log.debug("Fixed fn: [%s]" % rfn)


        if subdir is not None:
            rfn = os.path.join(subdir,rfn)

        if targetdir is None:
            fn = rfn
        else:
            fn = os.path.join(targetdir,rfn)
        
        with repo.openUrl(info.filename) as fh:
                
            lm = fh.info().get("Last-Modified")
                
            hd = info.checksums[repo.preferred_hash].saveAndCheckFile(fh,fn)
                
        log.info("Saved file [%s] with checksum [%s]"%(rfn,hd))

        if lm is not None:
            
            dt = util.parseRFC822Date(lm)
            log.info("Setting modification time of file [%s] to [%s]"%(rfn,dt))
            os.utime(fn,(datetime.now().timestamp(),dt.timestamp()))
            
        files.append(rfn)
    
    return files
