'''
  $Id$

  clazzes.org PBA package downloader

  (c)2018 clazzes.org project
  
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at
 
      http://www.apache.org/licenses/LICENSE-2.0
 
  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.

'''

import io
import hashlib
import email.utils
from time import mktime
from datetime import datetime

class RepoDataSupport:
    '''
       A helper class for data classes with python slots, which
       may be filled with None instances, if not read.
       Moreover, a verbose string representation for unit tests is provided.sssss
    '''
    def fillMissingSlots(self):
        for x in type(self).__slots__:
            if not hasattr(self,x):
                setattr(self,x,None)
                
    def __str__(self):
        t = type(self)
        with io.StringIO() as ih:
            ih.write(t.__name__)
            
            first=True
            for x in t.__slots__:
                ih.write('[' if first else ',')
                ih.write(x)
                ih.write('=')
                v = getattr(self,x)
                
                if isinstance(v,dict):
                    ih.write('{')
                    vfirst = True
                    for k in sorted(v.keys()):
                        if not vfirst:
                            ih.write(',')
                        ih.write(k)
                        ih.write('=')
                        ih.write(str(v[k]))
                        vfirst = False
                    ih.write('}')
                else:
                    ih.write(str(v))
                first=False
                
            ih.write(']')
            return ih.getvalue()
        
def getDomText(nodelist):
    rc = []
    for node in nodelist:
        if node.nodeType == node.TEXT_NODE:
            rc.append(node.data)
    return ''.join(rc)

def parseRFC822Date(rfc822date):
    '''
    Parse an RFC822 date like 'Wed, 07 Jun 2017 08:48:20 UTC'.
    :param rfc822date: A date as in HTTP headers and the like-
    :return: A datetime object. 
    '''
    date_parsed = email.utils.parsedate(rfc822date)
    return datetime.fromtimestamp(mktime(date_parsed))
                    

class WrongSignature(Exception):
    pass

class FilesizeMismatch(Exception):
    pass

class ChecksumMismatch(Exception):
    pass

class Checksum:
    '''
       A checksum with a given hex digest value and a digest type like 'sha256'.
    '''
    
    __slots__ = ('value','algorithm','size')
    
    def __init__(self,node=None,value=None,algorithm=None,size=None):
        '''
           :param node: An XML to initialize from, if None, initialize from value and algorithm parameters.
           :param value: The exptected hex digest.
           :param algorithm: The digest algorithm. 
           :param size: The expected file size or None if unknown 
        '''
        
        if node is not None:
            self.algorithm = node.attributes['type'].value
            self.value = getDomText(node.childNodes)
            self.size = None
        else:
            self.value = value
            self.algorithm = algorithm
            self.size = size
    
    def checkData(self,data,source):
        '''
           Check a given byte array against this checksum.
           :param data: The payload to check.
           :param source: The name of the underlying resource of the payload auch as a filename or an URL.
           :return: The hex digest of the provided data. 
        '''
        
        if self.size is not None and len(data) != self.size:
            raise ChecksumMismatch("Expected size resource [%s] differs (expected: %s, actual: %s)"%
                                   (source,self.size,len(data)))
        
        h = hashlib.new(self.algorithm)
        h.update(data)
        hd = h.hexdigest()
        if hd != self.value:
            raise ChecksumMismatch("Checksum for resource [%s] differs (expected: %s, actual: %s)"%
                                   (source,self.value,hd))
    
    def saveAndCheckFile(self,fh,fn):
        '''
           Read through a provided file-like handle, save it to a file and
           simultaneously check the contents against this checksum.
           :param fh: The file-like input handle.
           :param fn: The filename to save to.
           :return: The hex digest of the provided input. 
        '''

        h = hashlib.new(self.algorithm)

        n = 0

        with io.open(fn,"wb") as oh:
            data = fh.read(16384)
                    
            while len(data) > 0:
                
                n += len(data)
                oh.write(data)
                h.update(data)
                data = fh.read(16384)
        
        if self.size is not None and n != self.size:
            raise ChecksumMismatch("Expected size resource [%s] differs (expected: %s, actual: %s)"%
                                   (fn,self.size,n))
  
        hd = h.hexdigest()

        if hd != self.value:
            raise ChecksumMismatch("Checksum for file [%s] differs (expected: %s, actual: %s)"%
                                   (fn,self.value,hd))
        
        return hd
    
    def __str__(self):
        return "%s(%s)"%(self.algorithm,self.value)
