'''
  $Id$

  clazzes.org PBA package downloader

  (c)2018 clazzes.org project
  
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at
 
      http://www.apache.org/licenses/LICENSE-2.0
 
  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.

'''

import io
import os.path
import logging
import configparser
import tempfile
import sqlite3
import bz2
import gzip

from urllib.request import urlopen
from urllib.parse import urlparse, urljoin

from xml.dom import pulldom

from clazzes.pbadownload import util

log = logging.getLogger(__name__)

#
# rpm -D 'keyringpath /home/wglas/tmp/pubkeys' -K fancylog-1-1.0.3-1.x86_64.rpm 
# ls -l /home/wglas/tmp/pubkeys/
# insgesamt 4
# -rw-r--r-- 1 wglas wglas 1714 Feb  9  2016 RPM-GPG-KEY-pba-builder-iteg-at.key
#

class PackageEntry(util.RepoDataSupport):
    '''
      An entry in the packages table of an SQL lite database.
    '''
    __slots__ = ('name','arch', 'epoch','version','release','summary','size_package','size_installed','location','checksum')
    
    def __init__(self,row):
        self.name = row[0]
        self.arch = row[1]
        self.epoch = row[2]
        self.version = row[3]
        self.release = row[4]
        self.summary = row[5]
        self.size_package = row[6]
        self.size_installed = row[7]
        self.location = row[8]
        self.checksum = util.Checksum(value=row[9],algorithm=row[10])

class RepoDb:
    '''
      An opened SQLLite database of a repository, most likely the primary DB.
      This class is a context manage and holds a temporary directory as well an
      SQLLite DB connection.
    '''
    
    def __init__(self,data,name):
        
        self.name = name
        
        try:
            self.tmpdir = tempfile.TemporaryDirectory("pbadownload")
            
            fn = os.path.join(self.tmpdir.name,self.name)
            
            log.info ("Creating sqlite DB under [%s]"%fn)

            with io.open(fn,"wb") as fh:
                fh.write(data)
            
            self.conn = sqlite3.connect(fn)  # @UndefinedVariable
        
        except:
            self.tmpdir.cleanup()
    
    def findPackage(self,name):
        
        cur = self.conn.cursor()
        cur.execute("SELECT name,arch,epoch,version,release,summary,size_package,size_installed,location_href,pkgId,checksum_type FROM packages WHERE name=?",(name,))

        ret = []

        while True:
            row = cur.fetchone()
        
            if row is None:
                return ret
        
            ret.append(PackageEntry(row))
         
    def __enter__(self):
        return self
    
    def __exit__(self, exc, value, tb):  # @UnusedVariable
        try:
            self.conn.close()
        except:
            log.exception("Error closing SQL lite DB [%s]"%self.name)
        self.tmpdir.cleanup()

class MdEntry(util.RepoDataSupport):
    '''
       An entry in a repomd.xml file.
    '''
    
    __slots__ = ('checksum','database_version','location','open_size','open_checksum','size','timestamp','type')
    
    def __init__(self,node):
        
        if log.getEffectiveLevel() <= logging.DEBUG:
            log.debug("Initializing MdEntry from dom node [%s]"%(node.toxml(),))
        
        self.type = node.attributes['type'].value
        
        child = node.firstChild
        
        while child is not None:
            
            if child.localName == 'location':
                self.location = child.attributes['href'].value
            elif child.localName == 'checksum':
                self.checksum = util.Checksum(node=child)
            elif child.localName == 'open-checksum':
                self.open_checksum = util.Checksum(node=child)
            elif child.localName == 'size':
                self.size = int(util.getDomText(child.childNodes))
            elif child.localName == 'open-size':
                self.open_size = int(util.getDomText(child.childNodes))
            elif child.localName == 'timestamp':
                self.timestamp = int(util.getDomText(child.childNodes))
            elif child.localName == 'database_version':
                self.database_version = int(util.getDomText(child.childNodes))
            elif child.nodeType != child.TEXT_NODE:
                log.warn("repomd.xml entry contains invalid child tag [%s]"%child.localName)    
            
            child = child.nextSibling

        self.fillMissingSlots()


class RepoMd(util.RepoDataSupport):
    '''
       The parsed content of a repodata/repomd.xml file with the possibility to
       open a file list database in sql lite format.
    '''
    
    __slots__ = ('baseurl','entries','revision')
    
    def openUrl(self,relurl):
        '''
           Open an URL relative to the base URL of the repository.
           
           :param relurl: A relive URL not starting with a slash ant not containing '://' 
           :return: A file-like object as returned by urllib.request.urlopen().
        '''
        
        if relurl[0] == '/' or '://' in relurl:
            raise ValueError("Relative URL [%s] is absolute or contains a scheme.")
        
        url = self.baseurl + relurl
        
        log.info("Loading URL [%s]..."%url)
        
        return urlopen(url)
    
    def __init__(self,baseurl):
        
        if baseurl[-1] == '/':
            self.baseurl = baseurl
        else:
            self.baseurl = baseurl + '/'
    
        self.entries = {}
        depth = 0
        
        relurl = 'repodata/repomd.xml'
        
        with self.openUrl(relurl) as fh:
            doc = pulldom.parse(fh)
            for event, node in doc:
                if event == pulldom.START_ELEMENT:
                    if depth == 1 and node.tagName == 'data':
                        doc.expandNode(node)
                        entry = MdEntry(node)
                        self.entries[entry.type] = entry
                    elif depth == 1 and node.tagName == 'revision':
                        doc.expandNode(node)
                        self.revision = int(util.getDomText(node.childNodes))
                    elif depth == 0 and node.tagName == 'repomd':
                        depth = 1
                    else:
                        raise ValueError("repomd.xml from [%s] contains an invalid tag [%s] at depth [%d]"%(self.baseurl+relurl,node.tagName,depth))
    
        self.fillMissingSlots()
        log.info("Loaded repomd.xml with revision [%d] and [%d] entries."%(self.revision,len(self.entries)))

    def openDb(self,name='primary_db'):
        '''
        Open an SQL Lite DB contained in the repository.
        :param name: The name of an entry, which is 'primary_db' by default.
        :return:A context manager of type RepoDb for querying individual files. 
        '''
        
        entry = self.entries[name]
        
        with self.openUrl(entry.location) as fh:
            bzdata = fh.read()
            
        entry.checksum.checkData(bzdata,entry.location)
        
        if entry.location[-4:] == '.bz2':
            
            data = bz2.decompress(bzdata)
        
            entry.open_checksum.checkData(data,entry.location[0:-4])

        elif entry.location[-3:] == '.gz':
            
            data = gzip.decompress(bzdata)
        
            entry.open_checksum.checkData(data,entry.location[0:-3])

        else:
            data = bzdata

        return RepoDb(data,name)
        

class YumRepo(util.RepoDataSupport):
    '''
       The parsed repository description of a *.repo file loaded from an absolute URL.
    '''
    
    __slots__ = ('name','baseurl','gpgkey','gpgkeyurl','enabled','gpgcheck')
    
    def __init__(self,name,repourl):
        '''
          Load a yum repo description from an URL containing a repo file.
        '''
        if repourl[-1] == "/":
            self.repourl = None
            # baseurl directly specified, see PBA-35
            self.name = name
            self.baseurl = repourl
            
            log.info("Accessing yum repository with name [%s] and provided baseurl [%s]."%(self.name,self.baseurl))
            self.gpgkey=None
            self.gpgkeyurl=None
            self.enabled=True
            self.gpgcheck=False
        else:
            self.repourl = urlparse(repourl)
        
            log.info("Loading yum repository [%s] from [%s]"%(name,self.repourl.geturl()))
        
            with urlopen(repourl) as fh,io.TextIOWrapper(fh,encoding='utf-8') as th :
                config = configparser.ConfigParser()
                config.read_file(th,repourl)
                secs = config.sections()
                if len(secs) != 1:
                    raise ValueError("repo file from [%s] does not contain a single section"%repourl)
                
                secname = secs[0] 
                values = config[secname]
                
                self.name = values['name']
                if self.name != secname:
                    log.warn("Section name [%s] of repo file [%s] differs from name property [%s]"%
                             (secname,repourl,self.name))
                
                self.baseurl = values['baseurl']
                self.gpgkey = values.get('gpgkey')
                    
                if self.gpgkey is None:
                    self.gpgkeyurl = None
                else:
                    gpgkeyparsed = urlparse(self.gpgkey)
                    
                    if gpgkeyparsed.scheme == 'file' and self.repourl.scheme != 'file':
                        
                        self.gpgkeyurl = urljoin(repourl,'/gpg/'+gpgkeyparsed.path.split('/')[-1])
                        log.info("Will try to fetch GPG key [%s] from URL [%s]"%(self.gpgkey,self.gpgkeyurl))
                    else:
                        self.gpgkeyurl = None
                        
                self.enabled = int(values['enabled'])
                self.gpgcheck = int(values['gpgcheck'])
                        
                log.info("Successfully loaded yum repository with name [%s] and baseurl [%s]."%(self.name,self.baseurl))
                            
    def resolveBaseUrl(self,**kwargs):
        '''
          Resolve a base URL by replacing the placeholders specified as kwargs.
          Usual kwarags a 'releasever' and 'basearch' for resolving the distribution and architecture. 
        '''
        
        url = self.baseurl
        
        for k,v in kwargs.items():
            
            url = url.replace('$'+k,v)
            
        return url
